\ WebDAV filer
\ Source text for DX-Forth.
\ Version 1.0
\ Pavel Zampach, 2024
\ Intended for RomWBW and WiFi modem.
\ Revise data in WLIB.F!
\ for 30 rows screen, depended lines are marked by '#####'
\ *

inlib fileio
inlib statmem
inlib strcon
inlib strvar
inlib buflib
inlib arrlib
inlib applib
inlib modemrcv
inlib wlib
inlib ndr
cr ." WFile 1.0 loading"

decimal

104 constant FN-MAX			\ max. # of filenames#####
13 constant FN-LEN			\ length of filename
fn-max fn-len array: FN()		\ array of filenames
80 buffer: RBUF				\ response buffer
variable POINTER			\ pointer to FN()
integer #FN				\ # of filenames

\ --- Support words ---

s" \20HTTP/1.1" constr HTTP

\ Return substring [addr2 u2] from string [addr1 u1] delimited by char [c1]
\ (begin) and [c2] (end). Delimiter chars are not included.
: /QUOT   ( addr1 u1 c1 c2 -- addr2 u2 )
  >r scan 1 /string
  over swap r> scan drop over -
;

\ Type string [addr u] with Z-System marks (^A, ^B)
: ZTYPE   ( addr u -- )
  bounds do
    i c@ cond
      1 of bold endof			\ ^A
      2 of normal endof			\ ^B
      emit
    cont
  loop
;

: CONNECT   ( -- )
  flushport
  mreset mquiet
  web-url mdial flushport
;

: DISCONNECT   ( -- )
  s" +++" mcomflush
  s" ATH" mcomflush
  mreset
;

\ Extended file check
:NONAME   ( ior -- )
  if
    cr ." File access error" cr
    disconnect
    abort
  then
; is fc

: READ-RESPONSE   ( -- )
  500 to wticks
  rbuf rcv-line				\ the first line to RBUF
  20 to wticks
  begin					\ waiting for empty line (CRLF only)
    pad rcv-line
    pad c@ 2 =
  until
;

\ [flag] is TRUE if response in RBUF contents number [u]
: TEST-RESPONSE   ( u -- flag )
  rbuf count rot (u.) search nip nip
;

\ Response check
: RC   ( ior -- )
  if
    ." OK" cr
  else 
    cr pad count type cr
    disconnect
    abort
  then
;

\ [adr u] is actual filename
: POINT-FN   ( -- addr u )
  pointer @ fn() count
;

\ Closing part of HTML headers
: END-HEADER   ( -- )
  s" Host: localhost" send-line
  s" Authorization: Basic " >pad usr:psw apps send-line
  s" Connection: keep-alive" send-line
  send-empty
;

integer K
fn-len buffer: X

\ Move filename item from [addr1] to [addr2]
: FN-MOVE   ( addr1 addr2 -- )
  fn-len cmove
;

\ Compare filename
: COMP-N   ( addr1 addr2 -- flag )
  >r count r> count compare 0<
;

\ Sort filename array (selection sort algorithm)
: SORT   ( -- )
  #fn 1- 0 ?do				\ for i:=1 to n-1
    i to k				\ k:=i
    i fn() x fn-move			\ x:=a[i]
    #fn i 1+ do				\ for j:=i+1 to n
      i fn() x comp-n if		\ if a[j].key < x.key
        i to k				\ k:=j
        i fn() x fn-move		\ x:=a[j]
      then
    loop
    i fn() k fn() fn-move 		\ a[k]:=a[i]
    x i fn() fn-move			\ a[i]:=x
  loop
;

\ --- Read files ---

s" <?xml version=\221.0\22 encoding=\22utf-8\22?>" constr LINE1
s" <D:propfind xmlns:D='DAV:'>" constr LINE2
s" <D:prop><D:displayname/></D:prop>" constr LINE3
s" </D:propfind>" constr LINE4
line1 nip 2+ line2 nip 2+ + line3 nip 2+ + line4 nip 2+ +
  constant DATALEN
s" B0:$$$.XML" constr WRK-NAME		\ work file in fast RAM disk

: READ-HEADER   ( -- )
  s" PROPFIND " >pad web-dir apps http apps send-line
  s" Content-Type: text/xml; charset=\22utf-8\22" send-line
  s" Content-Length: " >pad datalen appn send-line
  end-header
  line1 send-line
  line2 send-line
  line3 send-line
  line4 send-line
;

: PARSE-LINE   ( addr u -- )
  s" <D:href>" caps search if
    [char] > [char] < /quot
    [ web-dir nip ] literal /string
    dup #fn fn-max < and if
      #fn fn() place
      #fn 1+ to #fn
    end
  then
  2drop
;

\ Process dowloaded XML file
: PRO-XML   ( -- )
  0 to #fn
  wrk-name r/o open-file fc to fid
  begin
    pad 80 fid read-line fc
  while
    pad swap parse-line
  repeat   drop
  fid close-file fc
  wrk-name delete-file fc
  sort
;

\ Read and process WebDAV directory
: READ-FILES   ( -- )
  wrk-name open-data
  ." reading "
  read-header
  read-response
  rcv-data
  close-data
  207 test-response rc
  pro-xml
  pointer off
;

\ --- Download file ---

s" B0:" constr DEST-DRIVE		\ fast RAM disk as destination

: DOWN-HEADER   ( -- )
  s" GET " >pad web-dir apps point-fn apps http apps send-line
  s" Accept: */*" send-line
  end-header
;

: DOWN-FILE   ( -- )
  dest-drive >pad point-fn apps open-data
  ." download "
  down-header
  read-response
  rcv-data
  close-data
  200 test-response rc
;

\ --- Delete file ---

: DEL-HEADER   ( -- )
  s" DELETE " >pad web-dir apps point-fn apps http apps send-line
  end-header
;

: DEL-FILE   ( -- )
  ." delete "
  del-header
  read-response
  204 test-response rc
;

\ --- Upload file ---

128 constant BUFLEN
$80 constant BUF

\ Append double number [d] to string [addr u1].
: APPD   ( addr u1 d -- addr u2 )
  (d.) apps
;

\ Uploaded file length [d]
: F-LEN ( -- d )
  fid file-size fc
;

24 string: INPUT

\ Input filename
: ENTER-FN   ( -- )
  ." File :"
  rbuf dup 24 accept
  2dup upper input s!
;

: UP-HEADER   ( -- )
  s" PUT " >pad web-dir apps input s@ -path apps http apps send-line
  s" Content-Length: " >pad f-len appd send-line
  end-header
;

: SEND-DATA   ( -- )
  begin
    buf buflen fid read-file fc dup
  while
    led-on
    buf swap send-string
    led-off
  repeat   drop
;

: UP-FILE   ( -- )
  ." \20 upload "
  input s@ -ndr r/o open-file fc to fid
  up-header
  send-data
  fid close-file fc
  read-response flushport
  201 test-response 204 test-response or rc
;

\ -------------------------------

: .HEADER   ( -- )
  page ." WFile 1.0" 8 spaces
  s" e\01X\02it arrows \01D\02ownload \01U\02pload up\01C\02mdLine de\01L\02ete"
  ztype cr
;

\ Display filenames
: .FN   ( -- )
  cr
  #fn if
    #fn 0 do
      i fn() count >fcb drop 1+ dup
      4 spaces 8 type [char] . emit 8 + 3 type 4 spaces
    loop
  else
    ." No files"
  then
;

: INIT-SCREEN   ( -- )
  read-files
  .header .fn
;

\ Set cursor coords
: POINT-XY   ( -- )
  pointer @ dup
  3 and 20 *				\ X coord
  swap 2/ 2/ 2+				\ Y coord
  at-xy  
;

: TRIM-POINTER   ( -- )
  pointer @ dup
  0< if pointer off then
  #fn < not if #fn 1- pointer ! then
;

: MOVE-POINTER   ( c -- )
  case
    $014f c@ of -4 pointer +! endof
    $0150 c@ of  4 pointer +! endof
    $0151 c@ of  1 pointer +! endof
    $0152 c@ of -1 pointer +! endof
  endcase
  trim-pointer
;

: PRO-DIR   ( -- )
  begin
    point-xy ." -->"   0 29 at-xy	\ #####
    30 spaces 13 emit
    key upcase
    point-xy 3 spaces   0 29 at-xy	\ #####
    dup move-pointer
    #fn if
      dup [char] D = if down-file then
      dup [char] L = if del-file init-screen then
    then
    dup [char] U = if enter-fn up-file init-screen then
    dup [char] C = if up-file init-screen then
    [char] X =
  until
;

: MAIN   ( -- )
  cmdtail input s!
  connect
  init-screen
  pro-dir
  disconnect
;

[DEFINED] allocate [IF]
turnkey main wfile
[THEN]
